/*
 *  linux/arch/arm/mm/fault.c
 *
 *  Copyright (C) 1995  Linus Torvalds
 *  Copyright (c) 2009, Code Aurora Forum. All rights reserved.
 *  Modifications for ARM processor (c) 1995-2004 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/module.h>
#include <linux/signal.h>
#include <linux/mm.h>
#include <linux/hardirq.h>
#include <linux/init.h>
#include <linux/kprobes.h>
#include <linux/uaccess.h>
#include <linux/page-flags.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/tlbflush.h>
#include <linux/kallsyms.h>

#ifdef CONFIG_EMULATE_DOMAIN_MANAGER_V7
#include <asm/domain.h>
#endif 

#include "fault.h"

static int show_count = 0;
void *exception_sp = NULL;
void* addr_of_next_thread_info = &init_thread_info.flags; 
void* addr_of_prev_thread_info = &init_thread_info.flags; 
static int is_in_do_TestSP = 0;

#ifdef CONFIG_KPROBES
static inline int notify_page_fault(struct pt_regs *regs, unsigned int fsr)
{
	int ret = 0;

	if (!user_mode(regs)) {
		
		preempt_disable();
		if (kprobe_running() && kprobe_fault_handler(regs, fsr))
			ret = 1;
		preempt_enable();
	}

	return ret;
}
#else
static inline int notify_page_fault(struct pt_regs *regs, unsigned int fsr)
{
	return 0;
}
#endif


void show_pte(struct mm_struct *mm, unsigned long addr)
{
	pgd_t *pgd;

	if (!mm)
		mm = &init_mm;

	printk(KERN_ALERT "pgd = %p\n", mm->pgd);
	pgd = pgd_offset(mm, addr);
	printk(KERN_ALERT "[%08lx] *pgd=%08lx", addr, pgd_val(*pgd));

	do {
		pmd_t *pmd;
		pte_t *pte;

		if (pgd_none(*pgd))
			break;

		if (pgd_bad(*pgd)) {
			printk("(bad)");
			break;
		}

		pmd = pmd_offset(pgd, addr);
		if (PTRS_PER_PMD != 1)
			printk(", *pmd=%08lx", pmd_val(*pmd));

		if (pmd_none(*pmd))
			break;

		if (pmd_bad(*pmd)) {
			printk("(bad)");
			break;
		}

		
		if (PageHighMem(pfn_to_page(pmd_val(*pmd) >> PAGE_SHIFT)))
			break;

		pte = pte_offset_map(pmd, addr);
		printk(", *pte=%08lx", pte_val(*pte));
		printk(", *ppte=%08lx", pte_val(pte[-PTRS_PER_PTE]));
		pte_unmap(pte);
	} while(0);

	printk("\n");
}


static void
__do_kernel_fault(struct mm_struct *mm, unsigned long addr, unsigned int fsr,
		  struct pt_regs *regs)
{
	
	if (fixup_exception(regs))
		return;

	
	bust_spinlocks(1);
	printk(KERN_ALERT
		"Unable to handle kernel %s at virtual address %08lx\n",
		(addr < PAGE_SIZE) ? "NULL pointer dereference" :
		"paging request", addr);

	show_pte(mm, addr);
	die("Oops", regs, fsr);
	bust_spinlocks(0);
	do_exit(SIGKILL);
}


static void
__do_user_fault(struct task_struct *tsk, unsigned long addr,
		unsigned int fsr, unsigned int sig, int code,
		struct pt_regs *regs)
{
	struct siginfo si;

#ifdef CONFIG_DEBUG_USER
	if (user_debug & UDBG_SEGV) {
		printk(KERN_DEBUG "%s: unhandled page fault (%d) at 0x%08lx, code 0x%03x\n",
		       tsk->comm, sig, addr, fsr);
		show_pte(tsk->mm, addr);
		show_regs(regs);
	}
#endif


	tsk->thread.address = addr;
	tsk->thread.error_code = fsr;
	tsk->thread.trap_no = 14;
	si.si_signo = sig;
	si.si_errno = 0;
	si.si_code = code;
	si.si_addr = (void __user *)addr;
	force_sig_info(sig, &si, tsk);
}

void do_bad_area(unsigned long addr, unsigned int fsr, struct pt_regs *regs)
{
	struct task_struct *tsk = current;
	struct mm_struct *mm = tsk->active_mm;

	
	if (user_mode(regs))
		__do_user_fault(tsk, addr, fsr, SIGSEGV, SEGV_MAPERR, regs);
	else
		__do_kernel_fault(mm, addr, fsr, regs);
}

#define VM_FAULT_BADMAP		0x010000
#define VM_FAULT_BADACCESS	0x020000

static int
__do_page_fault(struct mm_struct *mm, unsigned long addr, unsigned int fsr,
		struct task_struct *tsk)
{
	struct vm_area_struct *vma;
	int fault, mask;

	vma = find_vma(mm, addr);
	fault = VM_FAULT_BADMAP;
	if (!vma)
		goto out;
	if (vma->vm_start > addr)
		goto check_stack;

	
good_area:
	if (fsr & (1 << 11)) 
		mask = VM_WRITE;
	else
		mask = VM_READ|VM_EXEC|VM_WRITE;

	fault = VM_FAULT_BADACCESS;
	if (!(vma->vm_flags & mask))
		goto out;

	
survive:
	fault = handle_mm_fault(mm, vma, addr & PAGE_MASK, fsr & (1 << 11));
	if (unlikely(fault & VM_FAULT_ERROR)) {
		if (fault & VM_FAULT_OOM)
			goto out_of_memory;
		else if (fault & VM_FAULT_SIGBUS)
			return fault;
		BUG();
	}
	if (fault & VM_FAULT_MAJOR)
		tsk->maj_flt++;
	else
		tsk->min_flt++;
	return fault;

out_of_memory:
	if (!is_global_init(tsk))
		goto out;

	
	up_read(&mm->mmap_sem);
	yield();
	down_read(&mm->mmap_sem);
	goto survive;

check_stack:
	if (vma->vm_flags & VM_GROWSDOWN && !expand_stack(vma, addr))
		goto good_area;
out:
	return fault;
}

static int __kprobes
do_page_fault(unsigned long addr, unsigned int fsr, struct pt_regs *regs)
{
	struct task_struct *tsk;
	struct mm_struct *mm;
	int fault, sig, code;

	if (notify_page_fault(regs, fsr))
		return 0;

	tsk = current;
	mm  = tsk->mm;

	
	if (in_atomic() || !mm)
		goto no_context;

	
	if (!down_read_trylock(&mm->mmap_sem)) {
		if (!user_mode(regs) && !search_exception_tables(regs->ARM_pc))
			goto no_context;
		down_read(&mm->mmap_sem);
	}

	fault = __do_page_fault(mm, addr, fsr, tsk);
	up_read(&mm->mmap_sem);

	
	if (likely(!(fault & (VM_FAULT_ERROR | VM_FAULT_BADMAP | VM_FAULT_BADACCESS))))
		return 0;

	
	if (!user_mode(regs))
		goto no_context;

	if (fault & VM_FAULT_OOM) {
		
		printk("VM: killing process %s\n", tsk->comm);
		do_group_exit(SIGKILL);
		return 0;
	}
	if (fault & VM_FAULT_SIGBUS) {
		
		sig = SIGBUS;
		code = BUS_ADRERR;
	} else {
		
		sig = SIGSEGV;
		code = fault == VM_FAULT_BADACCESS ?
			SEGV_ACCERR : SEGV_MAPERR;
	}

	__do_user_fault(tsk, addr, fsr, sig, code, regs);
	return 0;

no_context:
	__do_kernel_fault(mm, addr, fsr, regs);
	return 0;
}


static int __kprobes
do_translation_fault(unsigned long addr, unsigned int fsr,
		     struct pt_regs *regs)
{
	unsigned int index;
	pgd_t *pgd, *pgd_k;
	pmd_t *pmd, *pmd_k;
	pte_t *pte;

	if (addr < TASK_SIZE)
		return do_page_fault(addr, fsr, regs);

	index = pgd_index(addr);

	
	pgd = cpu_get_pgd() + index;
	pgd_k = init_mm.pgd + index;

	if (pgd_none(*pgd_k))
		goto bad_area;

	if (!pgd_present(*pgd))
		set_pgd(pgd, *pgd_k);

	pmd_k = pmd_offset(pgd_k, addr);
	pmd   = pmd_offset(pgd, addr);

	if (pmd_none(*pmd_k))
		goto bad_area;

	if ( (fsr & 0x7) == 0x7 ) {
		pte = pte_offset_kernel(pmd_k, addr);
		if (pte_none(*pte))
			goto bad_area;
	}

	copy_pmd(pmd, pmd_k);
	return 0;

bad_area:
	do_bad_area(addr, fsr, regs);
	return 0;
}


static int
do_sect_fault(unsigned long addr, unsigned int fsr, struct pt_regs *regs)
{
	do_bad_area(addr, fsr, regs);
	return 0;
}


static int
do_bad(unsigned long addr, unsigned int fsr, struct pt_regs *regs)
{
	return 1;
}

static int
do_imprecise_ext(unsigned long addr, unsigned int fsr, struct pt_regs *regs)
{
#ifdef CONFIG_ARCH_MSM_SCORPION
	unsigned int regval;
	static unsigned char flush_toggle;

	asm("mrc p15, 0, %0, c5, c1, 0\n" 
	    : "=r" (regval));
	if (regval == 0x2) {
		
		flush_toggle ^= 1;
		if (flush_toggle)
			asm("mcr p15, 0, %0, c7, c5, 0\n"
			    :
			    : "r" (regval)); 
		
		asm("mcr p15, 7, %0, c15, c0, 1\n"
		    :
		    : "r" (regval));
		
		regval = 0;
		asm("mcr p15, 0, %0, c5, c1, 0\n"
		    :
		    : "r" (regval));
		return 0;
	}
#endif

	return 1;
}

static struct fsr_info {
	int	(*fn)(unsigned long addr, unsigned int fsr, struct pt_regs *regs);
	int	sig;
	int	code;
	const char *name;
} fsr_info[] = {
	
	{ do_bad,		SIGSEGV, 0,		"vector exception"		   },
	{ do_bad,		SIGILL,	 BUS_ADRALN,	"alignment exception"		   },
	{ do_bad,		SIGKILL, 0,		"terminal exception"		   },
	{ do_bad,		SIGILL,	 BUS_ADRALN,	"alignment exception"		   },
	{ do_bad,		SIGBUS,	 0,		"external abort on linefetch"	   },
	{ do_translation_fault,	SIGSEGV, SEGV_MAPERR,	"section translation fault"	   },
	{ do_bad,		SIGBUS,	 0,		"external abort on linefetch"	   },
	{ do_page_fault,	SIGSEGV, SEGV_MAPERR,	"page translation fault"	   },
	{ do_bad,		SIGBUS,	 0,		"external abort on non-linefetch"  },
	{ do_bad,		SIGSEGV, SEGV_ACCERR,	"section domain fault"		   },
	{ do_bad,		SIGBUS,	 0,		"external abort on non-linefetch"  },
	{ do_bad,		SIGSEGV, SEGV_ACCERR,	"page domain fault"		   },
	{ do_bad,		SIGBUS,	 0,		"external abort on translation"	   },
	{ do_sect_fault,	SIGSEGV, SEGV_ACCERR,	"section permission fault"	   },
	{ do_bad,		SIGBUS,	 0,		"external abort on translation"	   },
	{ do_page_fault,	SIGSEGV, SEGV_ACCERR,	"page permission fault"		   },
	
	{ do_bad,		SIGBUS,  0,		"unknown 16"			   },
	{ do_bad,		SIGBUS,  0,		"unknown 17"			   },
	{ do_bad,		SIGBUS,  0,		"unknown 18"			   },
	{ do_bad,		SIGBUS,  0,		"unknown 19"			   },
	{ do_bad,		SIGBUS,  0,		"lock abort"			   }, 
	{ do_bad,		SIGBUS,  0,		"unknown 21"			   },
	{ do_imprecise_ext,	SIGBUS,  BUS_OBJERR,	"imprecise external abort"	   }, 
	{ do_bad,		SIGBUS,  0,		"unknown 23"			   },
	{ do_bad,		SIGBUS,  0,		"dcache parity error"		   }, 
	{ do_bad,		SIGBUS,  0,		"unknown 25"			   },
	{ do_bad,		SIGBUS,  0,		"unknown 26"			   },
	{ do_bad,		SIGBUS,  0,		"unknown 27"			   },
	{ do_bad,		SIGBUS,  0,		"unknown 28"			   },
	{ do_bad,		SIGBUS,  0,		"unknown 29"			   },
	{ do_bad,		SIGBUS,  0,		"unknown 30"			   },
	{ do_bad,		SIGBUS,  0,		"unknown 31"			   }
};

void __init
hook_fault_code(int nr, int (*fn)(unsigned long, unsigned int, struct pt_regs *),
		int sig, const char *name)
{
	if (nr >= 0 && nr < ARRAY_SIZE(fsr_info)) {
		fsr_info[nr].fn   = fn;
		fsr_info[nr].sig  = sig;
		fsr_info[nr].name = name;
	}
}


asmlinkage void __exception
do_DataAbort(unsigned long addr, unsigned int fsr, struct pt_regs *regs)
{
	const struct fsr_info *inf = fsr_info + (fsr & 15) + ((fsr & (1 << 10)) >> 6);
	struct siginfo info;

#ifdef CONFIG_EMULATE_DOMAIN_MANAGER_V7
	if (emulate_domain_manager_data_abort(fsr, addr))
		return;
#endif

	if (!inf->fn(addr, fsr, regs))
		return;

	printk(KERN_ALERT "Unhandled fault: %s (0x%03x) at 0x%08lx\n",
		inf->name, fsr, addr);

	info.si_signo = inf->sig;
	info.si_errno = 0;
	info.si_code  = inf->code;
	info.si_addr  = (void __user *)addr;
	arm_notify_die("Oops - unhandled data abort", regs, &info, fsr, 0);
}

asmlinkage void __exception
do_PrefetchAbort(unsigned long addr, unsigned int ifsr, struct pt_regs *regs)
{
	struct vm_area_struct *vma;
	struct siginfo info;

#ifdef CONFIG_EMULATE_DOMAIN_MANAGER_V7
	if (emulate_domain_manager_prefetch_abort(ifsr, addr))
		return;
#endif

	do_translation_fault(addr, ifsr, regs);

	
	if ((ifsr & 0x40D) == 0x00D) {
		vma = find_vma(current->mm, addr);
		if (vma && !(vma->vm_flags & VM_EXEC)) {
			info.si_signo = SIGSEGV;
			info.si_errno = 0;
			info.si_code  = SEGV_ACCERR;
			info.si_addr  = (void __user *)addr;
			arm_notify_die("Oops - execution permission fault",
				       regs, &info, ifsr, 0);
		}
	}
}

static int full_check_fp(unsigned int* ptr,unsigned int addr_of_next_frame)
{
	unsigned int fp,sp;
	for(;;)
	{		
		printk(KERN_ALERT "addr_of_next_frame[%x]\n",addr_of_next_frame);
		if(addr_of_next_frame % 4 != 0)
		{
			printk(KERN_ALERT "full_check_fp returns 0000\n");
			return 0;			
		}
		fp = *(unsigned int*)(addr_of_next_frame - 12);
		sp = *(unsigned int*)(addr_of_next_frame - 8);
		if(fp % 4 != 0 || sp % 4 !=0)
		{
			printk(KERN_ALERT "full_check_fp returns 0000\n");
			return 0;			
		}		
		if(fp == 0) 
		{
			if(sp >= addr_of_next_frame) 
			{
				if(addr_of_next_frame > (unsigned int)ptr+THREAD_SIZE-1024) 
				{
					printk(KERN_ALERT "full_check_fp returns 1\n");
					return 1;			
				}
				else
				{
					printk(KERN_ALERT "full_check_fp returns 00000\n");
					return 0;					
				}
			}
			else
			{
				printk(KERN_ALERT "full_check_fp returns 000\n");
				return 0;				
			}
		}
		else if( sp < fp && 
		fp <= (unsigned int)ptr+THREAD_SIZE && 
		sp > addr_of_next_frame ) 
		{
			addr_of_next_frame = fp;
		}
		else
		{
			printk(KERN_ALERT "full_check_fp returns 0\n");
			return 0;			
		}
	}
	printk(KERN_ALERT "full_check_fp returns 00\n");
	return 0;
}

static unsigned int find_possible_fp(unsigned int* ptr)
{
	int i;
	for(i=0;i<THREAD_SIZE/sizeof(unsigned int)-1;i++)
	{
		if( ptr[i] > ptr[i+1] && 
		(unsigned int)ptr+THREAD_SIZE >= ptr[i] && 
		(unsigned int)(&ptr[i+1]) < ptr[i+1] && 
		(ptr[i] % 4) == 0 && 
		(ptr[i+1] % 4) == 0 
		)
		{
			if(full_check_fp(ptr,
			ptr[i]))
			{
				return (unsigned int)ptr + i*sizeof(unsigned int) + 12;
			}
		}
	}
	return 0;
}

asmlinkage void __exception
do_TestSP(unsigned int sp,
unsigned int old_sp,
unsigned int ret_addr)
{
	
	unsigned int possible_fp;
	int i;
	unsigned int* stack_ptr = (unsigned int*)sp;

#if 0
	if( (stack_ptr[4] & 0x1F) == 0x17)
	{
		memcpy((char*)(((unsigned int)exception_sp >> 13) << 13),
		addr_of_next_thread_info,
		sizeof(struct thread_info));

		memcpy(addr_of_next_thread_info,
		(char*)(((unsigned int)exception_sp >> 13) << 13),
		sizeof(struct thread_info));	
	}
#endif

	if(unlikely(is_in_do_TestSP == 1))
	{
		printk(KERN_ALERT "re-enter do_TestSP\n");
		while(1);			
	}
	is_in_do_TestSP = 1;
	show_count++;
				
	if( unlikely((old_sp & ~(THREAD_SIZE - 1)) != ((unsigned long)addr_of_next_thread_info & ~(THREAD_SIZE - 1)) &&
	(old_sp & ~(THREAD_SIZE - 1)) != ((unsigned long)addr_of_prev_thread_info & ~(THREAD_SIZE - 1))))
	{

		memcpy((char*)(((unsigned int)exception_sp >> 13) << 13),
		addr_of_next_thread_info,
		sizeof(struct thread_info));

		printk(KERN_ALERT "fatal error : sp is corrupted. sp[0x%x]\n",old_sp);
		printk(KERN_ALERT "ret addr[0x%x]",ret_addr);
#ifdef CONFIG_KALLSYMS
		print_symbol("(%s) ", ret_addr);
#endif		
		printk(KERN_ALERT "\n");

#if 0
		asm volatile ("mrc p15, 0, %0, c6, c0, 2" : "=r" (ifar));
		printk(KERN_ALERT "ifar[0x%x]",ifar);
#ifdef CONFIG_KALLSYMS
		print_symbol("(%s) ", ifar);
#endif		
		printk(KERN_ALERT "\n");
#endif
	
		printk(KERN_ALERT "start to dump exception stack\n");
		for(i=0;i<5;i++)
		{
			printk(KERN_ALERT "stack_ptr[%d]=[0x%x] ", i,stack_ptr[i]);
#ifdef CONFIG_KALLSYMS
			if(i==1) print_symbol("(%s) ", stack_ptr[i]);
#endif	
			printk(KERN_ALERT "\n");
		}
		printk(KERN_ALERT "dump exception stack done\n");
		
		printk(KERN_ALERT "next addr[%p] prev addr[%p]\n",addr_of_next_thread_info,addr_of_prev_thread_info);
		
		printk(KERN_ALERT "start to dump thread stack\n");
		possible_fp = find_possible_fp((unsigned int*)addr_of_next_thread_info);
		printk(KERN_ALERT "possible_fp[%x]\n",possible_fp);
		if(possible_fp)
		{
			c_backtrace(possible_fp,SVC_MODE);
		}
		printk(KERN_ALERT "dump thread stack done\n");
		
		while(1);
		memcpy(addr_of_next_thread_info,
		(char*)(((unsigned int)exception_sp >> 13) << 13),
		sizeof(struct thread_info));		
	}
	else
	{
#if 0
		if(show_count % 10000 == 0 || show_count < 10)
		{
			memcpy((char*)(((unsigned int)exception_sp >> 13) << 13),
			addr_of_next_thread_info,
			sizeof(struct thread_info));
			
			printk(KERN_INFO "ok[%d]\n",show_count);		
		
			memcpy(addr_of_next_thread_info,
			(char*)(((unsigned int)exception_sp >> 13) << 13),
			sizeof(struct thread_info));			
		}
#endif
	}
	is_in_do_TestSP = 0;	
}
